package ch.usi.dag.jp2.dump.xml;

import java.io.BufferedOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.zip.GZIPOutputStream;

import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

import ch.usi.dag.jp2.dump.Dumper;
import ch.usi.dag.jp2.runtime.CCTNode;

/**
 * @author Andreas Sewe
 */
public class DynamicCallGraphDumper implements Dumper {
    
    private static final String DCG_XMLNS = "http://jp-profiler.origo.ethz.ch/xmlns/dynamic-call-graph";
    
    public DynamicCallGraphDumper(String outputFilePrefix) throws IOException {
        final FileOutputStream fos = new FileOutputStream(outputFilePrefix + ".dcg.xml.gz");
        final GZIPOutputStream gzfos = new GZIPOutputStream(fos);
        out = new OutputStreamWriter(new BufferedOutputStream(gzfos), "UTF-8");
    }
    
    private final OutputStreamWriter out;
    
    private XMLStreamWriter writer;
    
    public void dump(CCTNode topLevel) throws IOException {
        try {
            writer = XMLOutputFactory.newInstance().createXMLStreamWriter(out);
            writer.writeStartDocument("UTF-8", "1.0");
            writer.writeStartElement("dynamicCallGraph");
            writer.setDefaultNamespace(DCG_XMLNS);
            writer.writeDefaultNamespace(DCG_XMLNS);
            dumpAllMethods(topLevel);
            writer.writeEndDocument();
            writer.close();
        } catch (XMLStreamException e) {
            throw new IOException(e);
        } finally {
            out.flush();
            out.close();
        }
    }
    
    private void dumpAllMethods(CCTNode topLevel) throws XMLStreamException {
        final Map<String, List<CCTNode>> nodesByMid = collectMethodNodes(topLevel);
        for (String mid : nodesByMid.keySet()) {
            final List<CCTNode> nodesWithMid = nodesByMid.get(mid);
            writer.writeStartElement(DCG_XMLNS, "method");
            
            writer.writeAttribute("declaringClass", NodeHelper.extractDeclaringClass(mid));
            writer.writeAttribute("name", NodeHelper.extractName(mid));
            writer.writeAttribute("params", NodeHelper.join(NodeHelper.extractParams(mid), " "));
            writer.writeAttribute("return", NodeHelper.extractReturn(mid));
            
            writer.writeStartElement(DCG_XMLNS, "executionCount");
            writer.writeCharacters(Long.toString(NodeHelper.sumExecutionCount(nodesWithMid)));
            writer.writeEndElement();
            
            final SortedMap<Integer, Set<String>> callsiteToTarget = collectCallsites(nodesWithMid);
            for (Integer instruction : callsiteToTarget.keySet()) {
                writer.writeStartElement(DCG_XMLNS, "callsite");
                writer.writeAttribute("instruction", instruction.toString());
                
                for (String targetMid : callsiteToTarget.get(instruction)) {
                    writer.writeStartElement(DCG_XMLNS, "methodRef");
                    writer.writeAttribute("declaringClass", NodeHelper.extractDeclaringClass(targetMid));
                    writer.writeAttribute("name", NodeHelper.extractName(targetMid));
                    writer.writeAttribute("params", NodeHelper.join(NodeHelper.extractParams(targetMid), " "));
                    writer.writeAttribute("return", NodeHelper.extractReturn(targetMid));
                    
                    writer.writeStartElement(DCG_XMLNS, "executionCount");
                    writer.writeCharacters(Long.toString(collectCalleeExecutionCount(nodesWithMid, targetMid, instruction)));
                    writer.writeEndElement();
                    
                    writer.writeEndElement();
                }
                writer.writeEndElement();
            }
            
            final TreeMap<Integer, Long> startToBasicBlockExecution = collectBasicBlockExecutions(nodesWithMid);
            Map<Integer, Integer> startInstructionToEndInstruction = startInstructionToEndInstruction(nodesWithMid.get(0));
            
            for (int startInstruction : startToBasicBlockExecution.keySet()) {
                writer.writeStartElement(DCG_XMLNS, "basicBlock");
                writer.writeAttribute("startInstruction", Integer.toString(startInstruction));
                writer.writeAttribute("endInstruction", Integer.toString(startInstructionToEndInstruction.get(startInstruction)));
                
                writer.writeStartElement(DCG_XMLNS, "executionCount");
                writer.writeCharacters(Long.toString(startToBasicBlockExecution.get(startInstruction)));
                writer.writeEndElement();
                
                writer.writeEndElement();
            }
            
            writer.writeEndElement();
        }
    }
    
    private Map<Integer, Integer> startInstructionToEndInstruction(CCTNode node) {
        final Map<Integer, Integer> startInstructionToEndInstruction = new HashMap<Integer, Integer>();
        final int[] bbSizes = CCTNode.bbSizes.get(node.getMID());
        for (int bbIndex = 0, bbStart = 1; bbIndex < bbSizes.length; bbStart += bbSizes[bbIndex++])
            startInstructionToEndInstruction.put(bbStart, bbStart + bbSizes[bbIndex] - 1);
        return startInstructionToEndInstruction;
    }
    
    private TreeMap<Integer, Long> collectBasicBlockExecutions(List<CCTNode> nodes) {
        final TreeMap<Integer, Long> basicBlockExecutions = new TreeMap<Integer, Long>();
        
        for (CCTNode node : nodes) {
            final List<Integer> sizes = node.getBasicBlockSizes();
            final List<Long> executionCounts = node.getBasicBlockExecutionCounts();
            for (int bbIndex = 0, bbStart = 1; bbIndex < executionCounts.size(); bbStart += sizes.get(bbIndex++)) {
                basicBlockExecutions.put(bbStart, executionCounts.get(bbIndex) + (basicBlockExecutions.containsKey(bbStart) ? basicBlockExecutions.get(bbStart) : 0));
            }
        }
        
        return basicBlockExecutions;
    }
    
    /**
     * Collects (by MID) all nodes starting at a given root node (in
     * XPath-terms: descendants-or-self).
     * 
     * @param root the root node to start collecting from
     * @return all nodes starting at the given root node 
     */
    private Map<String, List<CCTNode>> collectMethodNodes(CCTNode root) {
        final Map<String, List<CCTNode>> nodesByMid = new HashMap<String, List<CCTNode>>();
        collectMethodNodes(root.getCallees(), nodesByMid);
        return nodesByMid;
    }
    
    private void collectMethodNodes(CCTNode n, Map<String, List<CCTNode>> accumulator) {
        if (n != null) {
            List<CCTNode> nodes;
            if ((nodes = accumulator.get(n.getMID())) == null)
                accumulator.put(n.getMID(), nodes = new ArrayList<CCTNode>());
            nodes.add(n);
            
            collectMethodNodes(n.getCallees(), accumulator);
            collectMethodNodes(n.getLeft(), accumulator);
            collectMethodNodes(n.getRight(), accumulator);
        }
    }
    
    /**
     * Assumes that all {@code callers} refer to the same method.
     */
    private SortedMap<Integer, Set<String>> collectCallsites(List<CCTNode> callers) {
        final SortedMap<Integer, Set<String>> callsiteToTarget = new TreeMap<Integer, Set<String>>();
        for (CCTNode caller : callers) {
            for (CCTNode callee : NodeHelper.collectCallees(caller)) {
                Set<String> targets;
                if ((targets = callsiteToTarget.get(callee.getBytecodeIndex())) == null)
                    callsiteToTarget.put(callee.getBytecodeIndex(), targets = new HashSet<String>());
                targets.add(callee.getMID());
            }
        }
        return callsiteToTarget;
    }
    
    private long collectCalleeExecutionCount(List<CCTNode> callers, String calleeMid, Integer instruction) {
        long calleeExecutionCount = 0;
        
        for (CCTNode caller : callers)
            for (CCTNode callee : NodeHelper.collectCallees(caller))
                if (callee.getMID().equals(calleeMid) && instruction.equals(callee.getBytecodeIndex()))
                    calleeExecutionCount += callee.getExecutionCount();
        
        return calleeExecutionCount;
    }
}
