package ch.usi.dag.jp2.dump.xml;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

import ch.usi.dag.jp2.runtime.CCTNode;

/**
 * @author Andreas Sewe
 */
public final class NodeHelper {
    
    private static String stripNativePrefix(String mid) {
        return mid.startsWith(NATIVE_PREFIX) ? mid.substring(NATIVE_PREFIX.length()) : mid;
    }
    
    private static final String NATIVE_PREFIX = "[NATIVE]";
    
    public static String extractDeclaringClass(String mid) {
        mid = stripNativePrefix(mid);
        return mid.substring(0, mid.lastIndexOf('.'));
    }
    
    public static String extractName(String mid) {
        mid = stripNativePrefix(mid);
        return mid.substring(mid.lastIndexOf('.') + 1, mid.indexOf('('));
    }
    
    public static List<String> extractParams(String mid) {
        mid = stripNativePrefix(mid);
        final String descriptors = mid.substring(mid.indexOf('(') + 1, mid.indexOf(')'));
        final List<String> params = new ArrayList<String>();
        int start = 0;
        
        while (start < descriptors.length()) {
            int dim = 0;
            while (descriptors.charAt(start + dim) == '[') dim++;
            
            final String param;
            if (descriptors.charAt(start + dim) == 'L')
                param = descriptors.substring(start, descriptors.indexOf(';', start) + 1);
            else
                param = descriptors.substring(start, start + dim + 1);
            
            start += param.length();
            params.add(param);
        }
        
        return params;
    }
    
    public static String extractReturn(String mid) {
        mid = stripNativePrefix(mid);
        return mid.substring(mid.indexOf(')') + 1);
    }
    
    static String join(Collection<?> s, String delimiter) {
        StringBuilder builder = new StringBuilder();
        Iterator iter = s.iterator();
        while (iter.hasNext()) {
            builder.append(iter.next());
            if (!iter.hasNext()) {
                break;
            }
            builder.append(delimiter);
        }
        return builder.toString();
    }
    
    public static long sumExecutionCount(final List<CCTNode> nodes) {
        long executionCount = 0;
        for (CCTNode n : nodes)
            executionCount += n.getExecutionCount();
        return executionCount;
    }
    
    /**
     * Collects all callees of a node.
     * 
     * @param caller the caller
     * @return the list of calles, sorted by callsite in ascending order
     */
    public static List<CCTNode> collectCallees(CCTNode caller) {
        List<CCTNode> callees = new ArrayList<CCTNode>();
        accumulateSiblings(caller.getCallees(), callees);
        Collections.sort(callees, new Comparator<CCTNode>() {
            
            public int compare(CCTNode lhs, CCTNode rhs) {
                
                return lhs.getBytecodeIndex() - rhs.getBytecodeIndex();
            }
        });
        return callees;
    }
    
    private static void accumulateSiblings(CCTNode n, List<CCTNode> accumulator) {
        if (n == null) return;
        
        accumulator.add(n);
        accumulateSiblings(n.getLeft(), accumulator);
        accumulateSiblings(n.getRight(), accumulator);
    }
    
    private NodeHelper() { throw new UnsupportedOperationException("Not meant to be instantiated"); }
}
