package ch.usi.dag.jp2.dump.xml;

import java.io.BufferedOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.ArrayDeque;
import java.util.Deque;
import java.util.Iterator;
import java.util.List;
import java.util.zip.GZIPOutputStream;

import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

import ch.usi.dag.jp2.dump.Dumper;
import ch.usi.dag.jp2.runtime.CCTNode;
import ch.usi.dag.jp2.runtime.CallingContextTreeNode;

public class XmlDumper implements Dumper {
    
    private final OutputStreamWriter out;
    
    private XMLStreamWriter writer;
    
    private Deque<CallingContextTreeNode> stack =
        new ArrayDeque<CallingContextTreeNode>();
    
    private static final String CCT_XMLNS = "http://jp-profiler.origo.ethz.ch/xmlns/calling-context-tree";
    
    public XmlDumper(String outputFilePrefix) throws IOException {
        final FileOutputStream fos = new FileOutputStream(outputFilePrefix + ".cct.xml.gz");
        final GZIPOutputStream gzfos = new GZIPOutputStream(fos);
        out = new OutputStreamWriter(new BufferedOutputStream(gzfos), "UTF-8");
    }
    
    public void dump(CCTNode topLevel) throws IOException {
        try {
            writer = XMLOutputFactory.newInstance().createXMLStreamWriter(out);
            writer.writeStartDocument("UTF-8", "1.0");
            writer.writeStartElement("callingContextTree");
            writer.setDefaultNamespace(CCT_XMLNS);
            writer.writeDefaultNamespace(CCT_XMLNS);
            for (CallingContextTreeNode n : topLevel.collectCallees()) {
                stack.push(n);
                writeMethodElement();
                stack.pop();
            }
            writer.writeEndDocument();
            writer.close();
        } catch (XMLStreamException e) {
            throw new IOException(e);
        } finally {
            out.flush();
            out.close();
        }
    }
    
    private void writeMethodElement() throws XMLStreamException {
        final CallingContextTreeNode n = stack.peek();
        
        writer.writeStartElement(CCT_XMLNS, "method");
        writer.writeAttribute("declaringClass", n.getDeclaringClass());
        writer.writeAttribute("name", n.getName());
        writer.writeAttribute("params", NodeHelper.join(n.getParams(), " "));
        writer.writeAttribute("return", n.getReturn());
        
        writer.writeStartElement(CCT_XMLNS, "executionCount");
        writer.writeCharacters(Long.toString(n.getExecutionCount()));
        writer.writeEndElement();
        
        writer.writeStartElement(CCT_XMLNS, "executedInstructions");
        writer.writeCharacters(Long.toString(n.getExecutedInstructions()));
        writer.writeEndElement();
        
        dumpCallSites(n.collectCallees());
        
        dumpBasicBlocks(n);
        
        writer.writeEndElement();
    }
    
    private void dumpCallSites(List<? extends CallingContextTreeNode> callsites) throws XMLStreamException {
        Iterator<? extends CallingContextTreeNode> iterator = callsites.iterator();
        CallingContextTreeNode previousCallsite = null, callsite = iterator.hasNext() ? iterator.next() : null;
        
        while (callsite != null) {
            if (previousCallsite == null || previousCallsite.getBytecodeIndex() != callsite.getBytecodeIndex()) {
                writer.writeStartElement(CCT_XMLNS, "callsite");
                writer.writeAttribute("instruction", Integer.toString(callsite.getBytecodeIndex()));
            }
            
            stack.push(callsite);
            writeMethodElement();
            stack.pop();
            
            previousCallsite = callsite;
            callsite = iterator.hasNext() ? iterator.next() : null;
            
            if (callsite == null || previousCallsite.getBytecodeIndex() != callsite.getBytecodeIndex()) {
                writer.writeEndElement();
            }
        }
    }
    
    private void dumpBasicBlocks(CallingContextTreeNode n) throws XMLStreamException {
        final List<Integer> basicBlockSizes = n.getBasicBlockSizes();
        final List<Long> basicBlockExecutionCounts = n.getBasicBlockExecutionCounts();
        
        for (int bbIndex = 0, bbStart = 1; bbIndex < basicBlockExecutionCounts.size(); bbStart += basicBlockSizes.get(bbIndex++)) {
            writer.writeStartElement(CCT_XMLNS, "basicBlock");
            writer.writeAttribute("startInstruction", Integer.toString(bbStart));
            writer.writeAttribute("endInstruction", Integer.toString(bbStart + basicBlockSizes.get(bbIndex) - 1));
            
            writer.writeStartElement(CCT_XMLNS, "executionCount");
            writer.writeCharacters(Long.toString(basicBlockExecutionCounts.get(bbIndex)));
            writer.writeEndElement();
            
            writer.writeEndElement();
        }
    }
}
