/**
 * Copyright (c) 2010 Aibek Sarimbekov, Philippe Moret, Walter Binder
 * 
 * Permission to copy, use, modify (only the included source files, modification
 * of the binaries is not permitted) the software is granted provided this
 * copyright notice appears in all copies.
 * 
 * This software is  provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 */
package ch.usi.dag.jp2.instrument;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import java.util.Arrays;
import java.util.concurrent.ConcurrentHashMap;

import org.objectweb.asm.Opcodes;
import org.objectweb.asm.tree.*;
import org.objectweb.asm.util.AbstractVisitor;

import ch.usi.dag.jborat.agent.Instrumentation;
import ch.usi.dag.jborat.runtime.DynamicBypass;
import ch.usi.dag.jborat.utils.ConfigurationReader;
import ch.usi.dag.jborat.utils.Namespace;
import ch.usi.dag.jp2.dump.Dumper;
import ch.usi.dag.jp2.dump.text.TextDumper;
import ch.usi.dag.jp2.runtime.CCTNode;

import uk.ac.glasgow.jsinger.nanopatterns.PatternSpotter;

/**
 * @author Aibek Sarimbekov
 * @author Philippe Moret
 * @author Walter Binder
 * @author Andreas Sewe
 */
public class AddInstrumentation implements Instrumentation {

    private static PatternSpotter spotter = new PatternSpotter();
    
    /**
     * All opcodes before which a call to {@code JP2Runtime#setIndex(int)}
     * should be inserted.
     */
    private Set<Integer> setIndexOpcodes = new TreeSet<Integer>();
    
    /**
     * Should the given method be instrumented?
     * 
     * Certain methods cannot be instrumented as they are, e.g.,
     * <code>native</code>.
     */
    private static boolean shouldInstrument(ClassNode clazz, MethodNode method) {
        if (clazz.name.equals("java/lang/Object") && method.name.equals("<init>"))
            return false;
        else if ((method.access & Opcodes.ACC_NATIVE) != 0)
            return false;
        else if ((method.access & Opcodes.ACC_ABSTRACT) != 0)
            return false;
        else
            return true;
    }
    
    public void initialize() {
        CCTNode.bbSizes = new ConcurrentHashMap<String, int[]>();

        // Wing: used to hold the methods and their basic block instructions
        CCTNode.bbInst = new ConcurrentHashMap<String, int[][]>();

        // Wing: used to hold the nano-patterns
        CCTNode.nanoPatterns = new ConcurrentHashMap<String, int[]>();

        // Wing: used to hold the source file names
        CCTNode.sourceFileNames = new ConcurrentHashMap<String, String>();
        
        initializeSetIndexOpcodes();
        setUpShutdownHook();
    }
    
    private void setUpShutdownHook() {
        boolean old = DynamicBypass.getAndSet();
        try {
            Runtime.getRuntime().addShutdownHook(new DumperThread());
        } finally {
            DynamicBypass.set(old);
        }
    }
    
    private void initializeSetIndexOpcodes() {
        final Namespace instructions = new Namespace(new ConfigurationReader("conf/instructions").readAsStrings());
        
        for (int opcode = 0; opcode < AbstractVisitor.OPCODES.length; opcode++) {
            if (instructions.contains(AbstractVisitor.OPCODES[opcode])) 
                setIndexOpcodes.add(opcode);
        }
    }
    
    @SuppressWarnings("unchecked")
    public void instrument(ClassNode clazz) throws Exception {
        for (MethodNode method : (List<MethodNode>) clazz.methods) {
            if (!shouldInstrument(clazz, method))
                continue;
            final int firstFreeIndex = method.maxLocals;
            final int secondFreeIndex = firstFreeIndex + 1;
            final int thirdFreeIndex = secondFreeIndex + 1;
            final BasicBlockAnalysis bba = new BasicBlockAnalysis();
            
            InsnList instructions = method.instructions;
            final AbstractInsnNode[] originalInstructions = instructions.toArray();
            
            bba.analyse(clazz, method);
            setBBsizes(clazz.name, method, bba, instructions);
            
            Iterator<AbstractInsnNode> myIterator = instructions.iterator();
            boolean foundNative=false;
            while(myIterator.hasNext()){
                AbstractInsnNode instruction = (AbstractInsnNode) myIterator.next();
                if (instruction.getType()==MethodInsnNode.METHOD_INSN ){
                    MethodInsnNode invokeMethod = (MethodInsnNode)instruction;
                    if(invokeMethod.name.contains("ch_dag_usi_jp2_"))
                        foundNative=true;
                }
            }
            
            final String mid = foundNative ? AddInstrumentation.printNativeMID(clazz, method) : AddInstrumentation.printMID(clazz, method);
            
            CCTNode.bbSizes.put(mid, bba.bbSize);

            // Wing: store the basic block instructions within their CCTNode
            CCTNode.bbInst.put(mid, bba.bbInst);

            // Wing: store the nano-pattern within the CCTNode
            CCTNode.nanoPatterns.put(mid, bba.nanoPattern);

            // Wing: store the source file name within the CCTNode
            // No need to store in bba
            if (clazz.sourceFile == null)
              CCTNode.sourceFileNames.put(mid, "unknown");
            else
              CCTNode.sourceFileNames.put(mid, clazz.sourceFile);
            
            InsnList newInstructions = new InsnList();
            final LabelNode startTryCatchBlockLabel = new LabelNode();
            final LabelNode endTryCatchBlockLabel = new LabelNode();
            
            if (isMainMethod(clazz, method))
                newInstructions.add(new MethodInsnNode(Opcodes.INVOKESTATIC, "ch/usi/dag/jp2/runtime/JP2Runtime", "enableProfiling", "()V"));
            
            newInstructions.add(new InsnNode(Opcodes.ACONST_NULL));
            newInstructions.add(new VarInsnNode(Opcodes.ASTORE, firstFreeIndex));
            
            newInstructions.add(createGetIndexInstruction());
            newInstructions.add(new VarInsnNode(Opcodes.ISTORE, thirdFreeIndex));
            
            newInstructions.add(startTryCatchBlockLabel);
            
            newInstructions.add(createGetCurrentNodeInstruction());
            newInstructions.add(new VarInsnNode(Opcodes.ASTORE, firstFreeIndex));
            
            newInstructions.add(new VarInsnNode(Opcodes.ALOAD, firstFreeIndex));
            newInstructions.add(new LdcInsnNode(mid));
            newInstructions.add(getShortestPushIntInstruction(bba.bbSize.length));
            newInstructions.add(createInvokeProfileInvocationInstruction());
            newInstructions.add(new VarInsnNode(Opcodes.ASTORE, secondFreeIndex));
            
            // Add bytecode index before instrumenting anything
            AddBytecodeIndex(instructions, secondFreeIndex, originalInstructions, thirdFreeIndex, secondFreeIndex);
            
            // Add updates to dynamic metrics
            int z=0;
            for (int h=0; h< bba.bBInstructions.length; ++h) {
                AbstractInsnNode a = bba.bBInstructions[h];
                if (a == null)
                    continue;
                
                InsnList newList = new InsnList();
                newList.add(new VarInsnNode(Opcodes.ALOAD,secondFreeIndex));
                //changed LDC to appropriate one   
                newList.add(getShortestPushIntInstruction(z));
                newList.add(createProfileBBInstruction());
                
                a =findTheRealInstruction(a,instructions);
                instructions.insertBefore(a, newList);	
                z++;
            }
            
            instructions.insert(newInstructions);
            
            Iterator j = instructions.iterator();
            
            while(j.hasNext()){
                AbstractInsnNode instruction = (AbstractInsnNode) j.next();
                int opcode = instruction.getOpcode();
                if(opcode>=Opcodes.IRETURN && opcode<= Opcodes.RETURN)
                {
                    newInstructions.clear();
                    newInstructions.add(new VarInsnNode(Opcodes.ALOAD,firstFreeIndex));
                    newInstructions.add(createSetCurrentNodeInstruction());
                    
                    newInstructions.add(new VarInsnNode(Opcodes.ILOAD,thirdFreeIndex));
                    newInstructions.add(createSetIndexInstruction());
                    
                    instructions.insert(instruction.getPrevious(),newInstructions);
                }
            }
            newInstructions.clear();
            LabelNode normalExecutionFinallyBlockLabel = new LabelNode();
            newInstructions.add(endTryCatchBlockLabel);
            method.tryCatchBlocks.add(new TryCatchBlockNode(startTryCatchBlockLabel, endTryCatchBlockLabel, endTryCatchBlockLabel, null));
            
            newInstructions.add(new VarInsnNode(Opcodes.ILOAD,thirdFreeIndex));
            newInstructions.add(createSetIndexInstruction());
            
            newInstructions.add(new VarInsnNode(Opcodes.ALOAD,firstFreeIndex));
            newInstructions.add(createSetCurrentNodeInstruction());
            
            if (isMainMethod(clazz, method))
                newInstructions.add(new MethodInsnNode(Opcodes.INVOKESTATIC, "ch/usi/dag/jp2/runtime/JP2Runtime", "disableProfiling", "()V"));
            
            newInstructions.add(new InsnNode(Opcodes.ATHROW));
            newInstructions.add(normalExecutionFinallyBlockLabel);
            instructions.add(newInstructions);
            
            method.maxStack += 6;
            method.maxLocals += 3;
        }
    }
    
    private static boolean isMainMethod(ClassNode clazz, MethodNode method) {
        return clazz.name.equals(System.getProperty("ch.usi.dag.jp2.countFromMainClass".replace('.', '/'))) && method.name.equals("main") && method.desc.equals("([Ljava/lang/String;)V");
    }
    
    private static MethodInsnNode createProfileBBInstruction() {
        return new MethodInsnNode(Opcodes.INVOKEVIRTUAL, "ch/usi/dag/jp2/runtime/CCTNode", "profileBB", "(I)V");
    }
    
    private static MethodInsnNode createInvokeProfileInvocationInstruction() {
        return new MethodInsnNode(Opcodes.INVOKESTATIC, "ch/usi/dag/jp2/runtime/JP2Runtime", "invokeProfileInvocation", "(Lch/usi/dag/jp2/runtime/CCTNode;Ljava/lang/String;I)Lch/usi/dag/jp2/runtime/CCTNode;");
    }
    
    private static MethodInsnNode createGetCurrentNodeInstruction() {
        return new MethodInsnNode(Opcodes.INVOKESTATIC, "ch/usi/dag/jp2/runtime/JP2Runtime", "getCurrentNode", "()Lch/usi/dag/jp2/runtime/CCTNode;");
    }
    
    private static MethodInsnNode createGetIndexInstruction() {
        return new MethodInsnNode(Opcodes.INVOKESTATIC, "ch/usi/dag/jp2/runtime/JP2Runtime", "getIndex", "()I");
    }
    
    private static MethodInsnNode createSetIndexInstruction() {
        return new MethodInsnNode(Opcodes.INVOKESTATIC, "ch/usi/dag/jp2/runtime/JP2Runtime", "setIndex", "(I)V");
    }
    
    private static MethodInsnNode createSetCurrentNodeInstruction() {
        return new MethodInsnNode(Opcodes.INVOKESTATIC, "ch/usi/dag/jp2/runtime/JP2Runtime", "setCurrentNode", "(Lch/usi/dag/jp2/runtime/CCTNode;)V");
    }
    
    private static MethodInsnNode createInvokeProfileInvocationForCurrentThreadInstruction() {
        return new MethodInsnNode(Opcodes.INVOKESTATIC, "ch/usi/dag/jp2/runtime/JP2Runtime", "invokeProfileInvocationForCurrentThread", "(Lch/usi/dag/jp2/runtime/CCTNode;)V");
    }
    
    private static MethodInsnNode createInvokeProfileInvocationForObjectConstructorInstruction() {
        return new MethodInsnNode(Opcodes.INVOKESTATIC, "ch/usi/dag/jp2/runtime/JP2Runtime", "invokeProfileInvocationForObjectConstructor", "(Lch/usi/dag/jp2/runtime/CCTNode;)V");
    }
    
    private static AbstractInsnNode findTheRealInstruction(AbstractInsnNode a, InsnList instructions) {
        int indexOfa = instructions.indexOf(a);
        
        while(a.getOpcode() == -1){
            if (indexOfa + 1 == instructions.size())
                break;
            
            indexOfa++;
            a=instructions.get(indexOfa);
        }
        return a;
    }
    
    private InsnList AddBytecodeIndex(InsnList instructions, int indexOfLocalCCTNode, AbstractInsnNode[] originalInstructions, int thirdFreeIndex, int secondFreeIndex){
        final Iterator j = instructions.iterator();
        
        while (j.hasNext()){
            final AbstractInsnNode instruction = (AbstractInsnNode) j.next();
            final int bytecodeIndex = findBytecodeIndex(instruction, instructions, originalInstructions);
            final InsnList newInstructions = new InsnList();
            
            try {
                if (setIndexOpcodes.contains(instruction.getOpcode())) {
                    newInstructions.clear();
                    newInstructions.add(getShortestPushIntInstruction(bytecodeIndex));
                    newInstructions.add(createSetIndexInstruction());
                    
                    if (isCallToObjectConstructor(instruction)) {
                        newInstructions.add(new VarInsnNode(Opcodes.ALOAD,secondFreeIndex));
                        newInstructions.add(createInvokeProfileInvocationForObjectConstructorInstruction());
                    } else if (isCallToCurrentThread(instruction)) {
                        newInstructions.add(new VarInsnNode(Opcodes.ALOAD, secondFreeIndex));
                        newInstructions.add(createInvokeProfileInvocationForCurrentThreadInstruction());
                    }
                    
                    if (instruction.getPrevious() == null) {
                        instructions.insert(newInstructions);
                    } else {
                        instructions.insert(instruction.getPrevious(), newInstructions);
                    }
                }
            } catch (Throwable e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        }
        
        return instructions;
    }
    
    public static String printNativeMID(ClassNode clazz, MethodNode method) {
        return "[NATIVE]"+ "L" + clazz.name + ";." + method.name + method.desc;
    }
    public static String printMID(ClassNode clazz, MethodNode method) {
        return "L" + clazz.name + ";." + method.name + method.desc;
    }
    private static boolean isCallToObjectConstructor(AbstractInsnNode instruction) {
        if (instruction.getType() == MethodInsnNode.METHOD_INSN) {
            MethodInsnNode invokeMethod = (MethodInsnNode) instruction;
            return invokeMethod.owner.equals("java/lang/Object") && invokeMethod.name.equals("<init>");
        } else {
            return false;
        }
    }
    
    private static boolean isCallToCurrentThread(AbstractInsnNode instruction) {
        if (instruction.getType() == MethodInsnNode.METHOD_INSN) {
            MethodInsnNode invokeMethod = (MethodInsnNode) instruction;
            return invokeMethod.owner.equals("java/lang/Thread") && invokeMethod.name.equals("currentThread");
        } else {
            return false;
        }
    }
    
    private static int findBytecodeIndex(AbstractInsnNode instruction, InsnList instructions, AbstractInsnNode[] originalInstructions) {
        int index = 1;
        
        for (int i = 0; i < originalInstructions.length; i++){
            AbstractInsnNode compare = (AbstractInsnNode) originalInstructions[i];
            int opcode = compare.getOpcode();
            if (compare == instruction)
                break;
            if (opcode != -1)
                index++;
        }
        
        return index;
    }
    
    private static void setBBsizes(String className, MethodNode method, BasicBlockAnalysis bba, InsnList instructions){
        int i = 0;
	
	// Wing : store the nano-pattern for this method
        bba.nanoPattern = spotter.getMethodNanoPatterns(className, method);

        //System.out.println(method.name + method.desc + " " + Arrays.toString(nanoPattern));
        
        for (int h=0; h< bba.bBInstructions.length; ++h) {
            AbstractInsnNode a = bba.bBInstructions[h];
            if (a == null)
                continue;
            
            bba.bbSize[i] = sizeOfBB(a, bba, instructions);
            // Wing: for each basic block, store their instructions
            bba.bbInst[i] = instOfBB(a, bba, instructions, bba.bbSize[i]); 
            i++;
        }
    }
    
    private static int sizeOfBB(AbstractInsnNode instruction, BasicBlockAnalysis bba, InsnList instructions){
        int size = 0, startIndex = 0, indexOfInstruction = 0;
        AbstractInsnNode compare=null;
        // Find the index of instruction in bba.bBInstructions
        for (int z = 0; z < bba.bBInstructions.length; z++){
            if (instruction == bba.bBInstructions[z]) {
                indexOfInstruction=z;
                break;
            }
        }
        if (bba.bBInstructions.length!=1){
            startIndex=instructions.indexOf(instruction);
            
            if (indexOfInstruction!=bba.bBInstructions.length-1) {
                compare = bba.bBInstructions[indexOfInstruction+1];
            } else {
                compare=instructions.getLast();
                if(compare.getOpcode()!=-1)
                    size++;
            }
        }
        
        for (int i = startIndex; i < instructions.size(); i++) {
            int opcode = instructions.get(i).getOpcode();
            
            if (compare != null && compare == instructions.get(i))
                break;
            
            if (opcode != -1)
                size++;
        }
        return size;
    }

    // Wing: This method is like the one above but stores every instruction that would have been counted
    private static int[] instOfBB(AbstractInsnNode instruction, BasicBlockAnalysis bba, InsnList instructions, int size){
        int startIndex = 0, indexOfInstruction = 0, counter = 0;
        int[] inst = new int[size];
        AbstractInsnNode compare=null;
        // Find the index of instruction in bba.bBInstructions
        for (int z = 0; z < bba.bBInstructions.length; z++){
            if (instruction == bba.bBInstructions[z]) {
                indexOfInstruction=z;
                break;
            }
        }
        if (bba.bBInstructions.length!=1){
            startIndex=instructions.indexOf(instruction);
            
            if (indexOfInstruction!=bba.bBInstructions.length-1) {
                compare = bba.bBInstructions[indexOfInstruction+1];
            } else {
                compare=instructions.getLast();
                if(compare.getOpcode()!=-1)
                {
                    inst[counter] = getTrueOpcode(compare);
                    counter++;
                }
            }
        }
        
        for (int i = startIndex; i < instructions.size(); i++) {
            int opcode = getTrueOpcode(instructions.get(i));
            
            if (compare != null && compare == instructions.get(i))
                break;
            
            if (opcode != -1)
            {
                inst[counter] = opcode;
                counter++;
            } 
        }
        return inst;
    }
    
    /**
     * Returns the shortest possible instruction that pushes the the given
     * <code>int</code> onto the operand stack.
     * 
     * @param i the <code>int</code> to push
     * @return One of <code>iconst_m1</code>, <code>iconst_1</code>,
     *  <code>iconst_2</code>, <code>iconst_3</code>, <code>iconst_4</code>,
     *  <code>iconst_5</code>, <code>bipush</code>, <code>sipush</code>, and
     *  <code>ldc</code>
     */
    private static AbstractInsnNode getShortestPushIntInstruction(int i) {
        if (-1 <= i && i <= 5)
            return new InsnNode(i + Opcodes.ICONST_0);
        if (Byte.MIN_VALUE <= i && i <= Byte.MAX_VALUE)
            return new IntInsnNode(Opcodes.BIPUSH, i);
        if (Short.MIN_VALUE <= i && i <= Short.MAX_VALUE)
            return new IntInsnNode(Opcodes.SIPUSH, i);
        else
            return new LdcInsnNode(i);
    }
    
    private final class DumperThread extends Thread {
        
        private final String outputFilePrefix = System.getProperty("ch.usi.dag.jp2.outputFilePrefix");
        
        public void run() {
            boolean old = DynamicBypass.getAndSet();
            try {
                if (outputFilePrefix != null) {
                    for (String dumperClassName : System.getProperty("ch.usi.dag.jp2.dumpers", TextDumper.class.getName()).split(":")) {
                        Class<? extends Dumper> dumperClass = (Class<? extends Dumper>) this.getClass().getClassLoader().loadClass(dumperClassName);
                        Dumper dumper = dumperClass.getConstructor(String.class).newInstance(outputFilePrefix);
                        System.err.println("Start dumping with " + dumperClass + "...");
                        long t0=System.nanoTime();
                        dumper.dump(CCTNode.getRoot());
                        long t1 = System.nanoTime();
                        System.err.println("Finished dumping in " + (t1 - t0) / 1000000 + " msec");
//                        System.err.println("Total number of nodes is: "+TextDumper.methodCounter);
//                        System.err.println("Number of unique nodes is: "+TextDumper.methods.size());
                        System.err.println("Number of executed Basic Blocks: "+TextDumper.bbCounts);
                        System.err.println("Number of executed bytecodes: "+TextDumper.bytecodes);
                    }
                }
            } catch (NoSuchMethodException e) {
                e.printStackTrace();
            } catch (IllegalArgumentException e) {
                e.printStackTrace();
            } catch (SecurityException e) {
                e.printStackTrace();
            } catch (InstantiationException e) {
                e.printStackTrace();
            } catch (IllegalAccessException e) {
                e.printStackTrace();
            } catch (InvocationTargetException e) {
                e.printStackTrace();
            } catch (IOException e) {
                e.printStackTrace();
            } catch (ClassNotFoundException e) {
                e.printStackTrace();
            } finally {
                DynamicBypass.set(old);
            }
        }
    }

    /**
     * Simple method that checks whether the opcode is one of the LOAD or STORE
     * instructions 
     */
    public static int getTrueOpcode(AbstractInsnNode insn)
    {
	int opcode = insn.getOpcode();
	int trueOpcode = opcode;
	if (opcode >= 21 && opcode <= 25 && ((VarInsnNode) insn).var >= 0 && ((VarInsnNode) insn).var <= 3)
	{
		trueOpcode = (opcode - 20) * 4 + 22 + ((VarInsnNode) insn).var;
	}
	if (opcode >= 54 && opcode <= 58 && ((VarInsnNode) insn).var >= 0 && ((VarInsnNode) insn).var <= 3)
	{
		trueOpcode = (opcode - 53) * 4 + 55 + ((VarInsnNode) insn).var;
	}
    	return trueOpcode;
    }
}
