/**
 * Copyright (c) 2010 Aibek Sarimbekov, Philippe Moret, Walter Binder
 * 
 * Permission to copy, use, modify (only the included source files, modification
 * of the binaries is not permitted) the software is granted provided this
 * copyright notice appears in all copies.
 * 
 * This software is  provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 */
package ch.usi.dag.jp2.instrument;

import org.objectweb.asm.Opcodes;
import org.objectweb.asm.tree.*;

import java.io.*;

import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 * @author Aibek Sarimbekov
 * @author Philippe Moret
 * @author Walter Binder
 */
public final class BasicBlockAnalysis {
    
    final Set<Integer> bBNumbers = new LinkedHashSet<Integer>();
    
    String className;
    
    private MethodNode methodNode;
    private InsnList il;
    public AbstractInsnNode[] bBInstructions;
    public int[] bbSize;
    public int[][] bbInst; // Wing: new 2d array to hold the instructions within each basic block
    public int[] nanoPattern; // Wing: array to hold the nano-pattern vector for the method
    
    public void analyse(ClassNode clazz, MethodNode methodNode) throws IOException {
        this.methodNode = methodNode;
        il = methodNode.instructions;
        className = clazz.name;
        
        findBasicBlocks(methodNode);
        getBBInstructions();
    }
    
    private void findBasicBlocks(MethodNode methodNode) {
        bBNumbers.clear();
        bBNumbers.add(0);
        final InsnList instructions = methodNode.instructions;
        AbstractInsnNode[] instructionsList = new AbstractInsnNode[instructions.size()];
        instructionsList = instructions.toArray();
        
        for (int i = 0; i < instructionsList.length; i++) {
            final AbstractInsnNode instruction = instructionsList[i];
            
            if (instruction.getOpcode() == -1)
                continue;
            
            if (endsBasicBlock(instruction))
                grabBasicBlockNumbers(instruction, i, instructions);
        }
    }
    
    private void grabBasicBlockNumbers(AbstractInsnNode instruction, int i, InsnList instructions) {
    	AbstractInsnNode nextInstruction = null;
        switch (instruction.getType()) {
        // Case covers the following opcodes:
        //   IFEQ, IFNE, IFLT, IFGE, IFGT, IFLE, IF_ICMPEQ, IF_ICMPNE,
        //   IF_ICMPLT, IF_ICMPGE, IF_ICMPGT, IF_ICMPLE, IF_ACMPEQ, IF_ACMPNE,
        //   GOTO, JSR, IFNULL, and IFNONNULL.
        case AbstractInsnNode.JUMP_INSN: {
            final LabelNode targetInstruction = ((JumpInsnNode) instruction).label;
            final int targetId = instructions.indexOf(targetInstruction);
            
            bBNumbers.add(targetId);

            if (i + 2 <= instructions.size()){
            	nextInstruction = instructions.get(i+1);
            	if(nextInstruction.getType()==AbstractInsnNode.LABEL && nextInstruction.getNext()==null)
            		;
            	else
            		bBNumbers.add(i + 1);
            }
            break;
        }
        // Case covers the following opcodes:
        //   LOOKUPSWITCH.
        case AbstractInsnNode.LOOKUPSWITCH_INSN: {
            final List keys = ((LookupSwitchInsnNode) instruction).keys;
            final List labels = ((LookupSwitchInsnNode) instruction).labels;
            
            for (int t = 0; t < keys.size(); t++) {
                final LabelNode targetInstruction = (LabelNode) labels.get(t);
                final int targetId = instructions.indexOf(targetInstruction);
                bBNumbers.add(targetId);
            }
            
            final LabelNode defaultTargetInstruction = ((LookupSwitchInsnNode) instruction).dflt;
            final int defaultTargetId = instructions.indexOf(defaultTargetInstruction);
            bBNumbers.add(defaultTargetId);
            
            break;
        }
        
        // Case covers the following opcodes:
        //   TABLESWITCH.
        case AbstractInsnNode.TABLESWITCH_INSN: {
            final int minKey = ((TableSwitchInsnNode) instruction).min;
            final List labels = ((TableSwitchInsnNode) instruction).labels;
            
            for (int t = 0; t < labels.size(); t++) {
                final LabelNode targetInstruction = (LabelNode) labels.get(t);
                final int targetId = instructions.indexOf(targetInstruction);
                bBNumbers.add(targetId);
            }
            
            final LabelNode defaultTargetInstruction = ((TableSwitchInsnNode) instruction).dflt;
            final int defaultTargetId = instructions.indexOf(defaultTargetInstruction);
            bBNumbers.add(defaultTargetId);
            
            break;
        }
        
        default:
            if (i + 2 < instructions.size())
                bBNumbers.add(i + 1);
            break;
        }
    }
    
    private void getBBInstructions() {
        int[] indexesOfBB = new int[bBNumbers.size()];
        Iterator j = bBNumbers.iterator();
        int i = 0;
        bBInstructions = new AbstractInsnNode[indexesOfBB.length];
        bbSize = new int[indexesOfBB.length];
        // Wing: initialise one dimension of 2d array to match the number of blocks
        bbInst = new int[indexesOfBB.length][];
        while (j.hasNext()) {
            indexesOfBB[i] = (Integer) j.next();
            i++;
        }
        sort(indexesOfBB);
        int k = 0;
        for (int m = 0; m < indexesOfBB.length; m++) {
            bBInstructions[k] = il.get(indexesOfBB[m]);
            k++;
        }
    }
    
    private void sort(int[] array) {
        boolean t = true;
        int buf;
        while (t) {
            t = false;
            for (int j = 0; j < array.length - 1; j++) {
                if (array[j] > array[j + 1]) {
                    buf = array[j];
                    array[j] = array[j + 1];
                    array[j + 1] = buf;
                    t = true;
                }
            }
        }
    }
    
    private static int sizeOfBB(AbstractInsnNode instruction, BasicBlockAnalysis bba, InsnList instructions) {
        int size = 0;
        int startIndex = 0;
        int indexOfInstruction = 0;
        AbstractInsnNode compare = null;
        // finding the index of instruction in bba.bBInstructions
        for (int z = 0; z < bba.bBInstructions.length; z++) {
            if (instruction == bba.bBInstructions[z]) {
                indexOfInstruction = z;
                break;
            }
        }
        if (bba.bBInstructions.length != 1) {
            startIndex = instructions.indexOf(instruction);
            
            if (indexOfInstruction != bba.bBInstructions.length - 1)
                compare = bba.bBInstructions[indexOfInstruction + 1];
            else {
                compare = instructions.getLast();
                if (compare.getOpcode() != -1)
                    size++;
            }
        }
        for (int i = startIndex; i < instructions.size(); i++) {
            int opcode = instructions.get(i).getOpcode();
            if (compare != null) {
                if (compare == instructions.get(i))
                    break;
            }
            if (opcode != -1)
                size++;
        }
        return size;
    }
    
    private static boolean endsBasicBlock(AbstractInsnNode instruction) {
        return
           instruction instanceof JumpInsnNode ||
           instruction instanceof LookupSwitchInsnNode ||
           instruction instanceof TableSwitchInsnNode ||
           instruction.getOpcode() == Opcodes.ATHROW ||
           instruction.getOpcode() == Opcodes.RET ||
           instruction.getOpcode() == Opcodes.JSR ||
           instruction.getOpcode() >= Opcodes.IRETURN && instruction.getOpcode() <= Opcodes.RETURN;
    }
}
