/**
 * Copyright (c) 2010 Aibek Sarimbekov, Philippe Moret, Walter Binder, Andreas
 * Sewe
 * 
 * Permission to copy, use, modify (only the included source files, modification
 * of the binaries is not permitted) the software is granted provided this
 * copyright notice appears in all copies.
 * 
 * This software is  provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 */
package ch.usi.dag.jp2.runtime;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLongArray;
import java.util.concurrent.atomic.AtomicLongFieldUpdater;
import java.util.concurrent.atomic.AtomicReferenceFieldUpdater;

import ch.usi.dag.jborat.runtime.DynamicBypass;

/**
 * A node in the calling context tree (CCT).
 * 
 * @author Aibek Sarimbekov
 * @author Philippe Moret
 * @author Walter Binder
 * @author Andreas Sewe
 */
public final class CCTNode implements CallingContextTreeNode {
    
    /**
     * The method identifier.
     */
    private final String mid;
    
    /**
     * The execution counts of this method's basic blocks.
     */
    private final AtomicLongArray basicBlockExecutionCounts;
    
    private final CCTNode parent; // Parent in the CCT
    
    private volatile CCTNode left, right; // Siblings in the CCT
    
    private volatile CCTNode callees; // Children in the CCT
    
    private final int bytecodeIndex; // Index of the call-site
    
    private volatile long executionCount; // Number of method executions
    
    private static CCTNode ROOT;
    
    private static final AtomicReferenceFieldUpdater<CCTNode, CCTNode> leftUpdater, rightUpdater;
    
    private static final AtomicReferenceFieldUpdater<CCTNode, CCTNode> calleesUpdater;
    
    private static final AtomicLongFieldUpdater<CCTNode>callsUpdater;
    
    public static ConcurrentHashMap<String, int[]> bbSizes;

    // Wing: ConcurrentHashMap used to store the instructions for each method's basic blocks
    public static ConcurrentHashMap<String, int[][]> bbInst;

    // Wing: ConcurrentHashMap used to store the nano-pattern for each method
    public static ConcurrentHashMap<String, int[]>  nanoPatterns;

    // Wing: ConcurrentHashMap used to store the source file name for each method
    public static ConcurrentHashMap<String, String> sourceFileNames;
    
    static {
        boolean old = DynamicBypass.getAndSet();
        try{
            leftUpdater =
                AtomicReferenceFieldUpdater.newUpdater(CCTNode.class, CCTNode.class, "left");
            rightUpdater =
                AtomicReferenceFieldUpdater.newUpdater(CCTNode.class, CCTNode.class, "right");
            calleesUpdater =
                AtomicReferenceFieldUpdater.newUpdater(CCTNode.class, CCTNode.class, "callees");
            callsUpdater = 
                AtomicLongFieldUpdater.newUpdater(CCTNode.class,"executionCount");
            ROOT  = new CCTNode(null, null, 0,-1,0);
        }
        finally{
            DynamicBypass.set(old);
        }
    }
    
    public static CCTNode getRoot() {
        return ROOT;
    }
    
    private CCTNode(CCTNode parent, String MID, int bytecodeIndex, int numerOfBB) {
        this(parent, MID, 1,bytecodeIndex, numerOfBB);
    }
    
    private CCTNode(CCTNode parent, String mid, int executionCount, int bytecodeIndex, int numberOfBB) {
        this.parent = parent;
        this.mid = mid;
        this.executionCount = executionCount;
        this.bytecodeIndex = bytecodeIndex;
        this.basicBlockExecutionCounts = new AtomicLongArray(numberOfBB);
    }
    
    /**
     * @return the parent node in the CCT (or <code>null</code> if this node it
     *   the root)
     */
    public CCTNode getParent() {
        return parent;
    }
    
    /**
     * @return the left sibling in the CCT (if any)
     */
    public CCTNode getLeft() {
        return left;
    }
    
    /**
     * @return the right sibling in the CCT (if any)
     */
    public CCTNode getRight() {
        return right;
    }
    
    public CCTNode getCallees() {
        return callees;
    }
    
    public String getMID() {
        return mid;
    }
    
    public int getBytecodeIndex(){
        return bytecodeIndex;
    }
    
    public long getExecutionCount() {
        return executionCount;
    }
    
    public long getExecutedInstructions() {
        final int[] bbSize = CCTNode.bbSizes.get(mid);
        long result = 0;
        
        for(int i = 0; i < basicBlockExecutionCounts.length(); i++)
            result += basicBlockExecutionCounts.get(i) * bbSize[i];
        
        return result;
    }
    
    public List<Integer> getBasicBlockSizes() {
        final int[] basicBlockSizes = CCTNode.bbSizes.get(mid);
        final List<Integer> result = new ArrayList<Integer>(basicBlockSizes.length);
        
        for (int i = 0; i < basicBlockSizes.length; i++)
            result.add(i, basicBlockSizes[i]);
        
        return result;
    }
    
    public List<Long> getBasicBlockExecutionCounts() {
        final List<Long> result = new ArrayList<Long>(basicBlockExecutionCounts.length());
        
        for (int i = 0; i < basicBlockExecutionCounts.length(); i++)
            result.add(i, basicBlockExecutionCounts.get(i));
        
        return result;
    }

    // Wing: This method returns the 2d array containing the basic block instructions
    public int[][] getBasicBlockInst() {
        return CCTNode.bbInst.get(mid);
    }

    // Wing: This method returns the array containing the nano-patterns
    public int[] getNanoPattern() {
        return CCTNode.nanoPatterns.get(mid);
    }

    // Wing: This method returns the source file name
    public String getSourceFileName() {
        return CCTNode.sourceFileNames.get(mid);
    }
    
    /**
     * 
     * @return the the array that contains counters for executed basic blocks as String
     */
    public String getBBCountsAsString(){
        return basicBlockExecutionCounts.toString();
    }
    
    /**
     * increments the counter for the executed basic block
     * 
     * @param bbIndex - index of the basic block in the method
     */
    public void profileBB(int bbIndex) {
        boolean old = DynamicBypass.getAndSet();
        try {
            if (Thread.currentThread().activateBytecodeCounting)
                basicBlockExecutionCounts.incrementAndGet(bbIndex);
        } finally {
            DynamicBypass.set(old);
        }
    }
    
    public CCTNode profileCall(String mid, int bytecodeIndex, int numberOfBB) { 
        CCTNode n, allocated;
        int initCalls=0;
        
        if (this == ROOT)
            bytecodeIndex=-1;
        
        if ((n = calleesUpdater.get(this)) == null) {
            if (Thread.currentThread().activateBytecodeCounting)
                initCalls=1;
            if (calleesUpdater.compareAndSet(this, null, (allocated = new CCTNode(this, mid,initCalls,bytecodeIndex, numberOfBB)))) {
                return allocated;
            } else {
                n = calleesUpdater.get(this);
            }
        }
        
        int hash_MID =  System.identityHashCode(mid); 
        
        while (true) {
            String n_MID;
            if ((n_MID = n.mid) == mid && n.bytecodeIndex==bytecodeIndex) {
                if (Thread.currentThread().activateBytecodeCounting)
                    callsUpdater.incrementAndGet(n);
                return n;
            } else if (hash_MID <= System.identityHashCode(n_MID)) { 
                CCTNode nLeft;
                if ((nLeft = leftUpdater.get(n)) == null) {
                    if (Thread.currentThread().activateBytecodeCounting)
                        initCalls=1;
                    if (leftUpdater.compareAndSet(n, null, (allocated = new CCTNode(this, mid, initCalls, bytecodeIndex, numberOfBB)))) {
                        return allocated;
                    } else {
                        n = leftUpdater.get(n);
                    }
                } else {
                    n = nLeft;
                }
            } else {
                CCTNode nRight;
                if ((nRight = rightUpdater.get(n)) == null) {
                    if (Thread.currentThread().activateBytecodeCounting)
                        initCalls=1;
                    if (rightUpdater.compareAndSet(n, null, (allocated = new CCTNode(this, mid, initCalls, bytecodeIndex, numberOfBB)))) {
                        return allocated;
                    } else {
                        n = rightUpdater.get(n);
                    }
                } else {
                    n = nRight;
                }
            } 
        }
    }
    
    public String getDeclaringClass() {
        final String strippedMid = stripNativePrefix(mid);
        return strippedMid.substring(0, strippedMid.lastIndexOf('.'));
    }
    
    public String getName() {
        final String strippedMid = stripNativePrefix(mid);
        return strippedMid.substring(strippedMid.lastIndexOf('.') + 1, strippedMid.indexOf('('));
    }
    
    public List<String> getParams() {
        final String strippedMid = stripNativePrefix(mid);
        final String descriptors = strippedMid.substring(strippedMid.indexOf('(') + 1, strippedMid.indexOf(')'));
        final List<String> params = new ArrayList<String>();
        int start = 0;
        
        while (start < descriptors.length()) {
            int dim = 0;
            while (descriptors.charAt(start + dim) == '[') dim++;
            
            final String param;
            if (descriptors.charAt(start + dim) == 'L')
                param = descriptors.substring(start, descriptors.indexOf(';', start) + 1);
            else
                param = descriptors.substring(start, start + dim + 1);
            
            start += param.length();
            params.add(param);
        }
        
        return params;
    }
    
    public String getReturn() {
        final String strippedMid = stripNativePrefix(mid);
        return strippedMid.substring(strippedMid.indexOf(')') + 1);
    }
    
    private static String stripNativePrefix(String mid) {
        return mid.startsWith(NATIVE_PREFIX) ? mid.substring(NATIVE_PREFIX.length()) : mid;
    }
    
    private static final String NATIVE_PREFIX = "[NATIVE]";
    
    public List<? extends CallingContextTreeNode> collectCallees() {
        List<CCTNode> callees = new ArrayList<CCTNode>();
        accumulateSiblings(this.getCallees(), callees);
        Collections.sort(callees, new Comparator<CCTNode>() {
            
            public int compare(CCTNode lhs, CCTNode rhs) {
                
                return lhs.getBytecodeIndex() - rhs.getBytecodeIndex();
            }
        });
        return callees;
    }
    
    private static void accumulateSiblings(CCTNode n, List<CCTNode> accumulator) {
        if (n == null) return;
        
        accumulator.add(n);
        accumulateSiblings(n.getLeft(), accumulator);
        accumulateSiblings(n.getRight(), accumulator);
    }
}
